
/*
 * struts.c : This test program reads link information from a VECT file and
 *            outputs the strut information in various (hopefully useful)
 *            formats. 
 */

/* Copyright 2004 The University of Georgia. */

/* This file is part of liboctrope.
   
liboctrope is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

liboctrope is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with liboctrope; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "octrope.h"
#include <argtable2.h>

#ifndef min 
  #define min(A,B) ((A < B) ? A : B)
#endif

struct edge_record {
  int comp;
  int lead_vert;
  int n_struts;
};

typedef struct augmented_strut_type {
    const octrope_strut *strut;
    double compression;
    double s,t;
    double end;
} augmented_strut;

typedef struct plot_box_type {
    int pos[2];
    int nstruts;    
} plot_box;

#define pline_edges(P) (((P).acyclic) ? (P).nv-1 : (P).nv)

int compare_augmented_struts(const void *a, const void *b)
{
  augmented_strut *A,*B;

  A = (augmented_strut *)(a);
  B = (augmented_strut *)(b);

  if (A->s < B->s) {
    
    return -1; 

  } else {

    return 1;

  }

}

int compare_mrlocs(const void *a, const void *b)
{
  octrope_mrloc *A,*B;

  A = (octrope_mrloc *)(a);
  B = (octrope_mrloc *)(b);

  if (A->component != B->component) {

    return A->component - B->component;

  } 

  if (A->vert != B->vert) {
    
    return A->vert - B->vert; 

  } 

  return 0;

}

void create_st_plot(const octrope_link *L, 
		    int n_struts, 
		    const octrope_strut *strutlist, 
		    int num_min_rad_locs,
		    octrope_mrloc *min_rad_locs,
		    FILE *outfile)

    /* Create a very informative plot showing curvature, struts, 
       and regions of interest such as straight segments for L 
       in arclength form. */

{ 
  int NBOXES;

  double tot_length,edgelength;
  int cmp,nv,vert,strut,i,j,k;
  int end;
  octrope_pline *cp;
  const octrope_strut *st;
  
  octrope_vector temp_vect;
  augmented_strut *as_list;
  
  double **s_positions;  
  double boxwidth, avg_edgelength;
 
  double plotscale;
  int slo, shi;
  double maxskip, s_skip;
  double straightlength;
  double tickplotstep,tick_s_step,tickpos;
  int NTICKS = {10};

  int intpos;
  double curv,curvpixel;
  int printedpt,pt_ok,printed_total;

  double yi,yj,mij,predicted_curvk;
  
  if (L == NULL) {
    return;
  }

  NBOXES = octrope_link_edges(L);

  printf("Building s,t plot for %d edge link with %d struts...\n\n",
	 NBOXES,n_struts);  

  /* First, compute length of link and s positions of verts. */  
  
  tot_length = 0;
  
  s_positions = (double **)(calloc(L->nc,sizeof(double *)));
  for (cmp = 0; cmp < L->nc; cmp++) {
    
    s_positions[cmp] = calloc(L->cp[cmp].nv,sizeof(double));   
    cp = &L->cp[cmp];
    nv = (cp->acyclic) ? cp->nv-1 : cp->nv;
    for (vert = 0; vert < nv; vert++) {
      s_positions[cmp][vert] = tot_length;  /* record s position of vertex */    
      
      temp_vect = cp->vt[vert+1];
      octrope_vsub(temp_vect,cp->vt[vert]);
      tot_length += octrope_norm(temp_vect);
                      
    }
  }
  
  boxwidth = tot_length/(double)(NBOXES);
  avg_edgelength = tot_length/(double)(octrope_link_edges(L));
  
  printf("Found arclength positions for vertices...\n");
 
  /* Now compute augmented strut list. We list every strut twice,
     switching endpoints to ensure */
  /* that we have both s and t coordinates in the augmented list. */
  
  as_list = malloc(2 *n_struts * sizeof(augmented_strut));
  
  for (strut=0;strut < n_struts;strut++) {
    
    st = &(strutlist[strut]);
    
    for (end=0;end<2;end++) {
  
      temp_vect = L->cp[st->component[end]].vt[st->lead_vert[end]+1];
      octrope_vsub(temp_vect,L->cp[st->component[end]].vt[st->lead_vert[end]]);
      edgelength = octrope_norm(temp_vect);
  
      as_list[2*strut + end].s = 
	s_positions[st->component[end]][st->lead_vert[end]] + st->position[end] * edgelength;
                   
      as_list[2*strut + end].end = end;
      as_list[2*strut + end].strut = st;
      
    }      
    
    as_list[2*strut].t = as_list[2*strut + 1].s;  /* Record t values */
    as_list[2*strut + 1].t = as_list[2*strut].s;
    
  }

  qsort(as_list,2*n_struts,sizeof(augmented_strut),compare_augmented_struts);
  printf("Built augmented strut list...\n");
   
  /* We are now ready to plot some boxes (!) and to identify
     potentially interesting regions of the graph
     automagically. First, we set up the graphics state for box
     plotting and have at it.*/
  
  tickplotstep = NBOXES*10.0/(double)(NTICKS);
  tick_s_step  = tot_length/(double)(NTICKS);

  fprintf(outfile,"%f %f translate \n",72*(3.5/2.0),72*3.0);    
  /* Center a 5"x5" box on the page */   

  plotscale = 5.0*72.0/(double)(10*NBOXES);
  /* Now one (PS) unit is 1/10 box */

  fprintf(outfile,"%f %f scale \n",plotscale,plotscale);  

  /* We now draw the outline box and grid. */

  fprintf(outfile,"/tls { 0.1 setgray %g setlinewidth } def\n",1.0/plotscale);
  fprintf(outfile,"tls\n");

  fprintf(outfile,
	  "0 0 moveto %d 0 rlineto "
	  "0 %d rlineto -%d -%d rlineto "
	  "0.9 setgray fill \n",
	  NBOXES*10,NBOXES*10,NBOXES*10,NBOXES*10);  

  /* And the same for the curvature plot below the main plot. */
  
  fprintf(outfile,"0.9 setgray 0 %g %d %g rectfill\n",
	  -2.5*tickplotstep,10*NBOXES,2*tickplotstep);

  /* The first step is to setup a procedure for drawing "label-plus-tick" */

  fprintf(outfile,"0.1 setgray\n");
  fprintf(outfile,"/TF /Helvetica findfont %g scalefont def\n",9.0/plotscale);
  fprintf(outfile,"/dt { %g %g moveto dup stringwidth pop neg 0 rmoveto show"
	  " newpath 0 0 moveto %g %g lineto stroke } def \n",
	  -6.0/plotscale,4.0/plotscale,-4.0/plotscale,4.0/plotscale); 

  /* Now drop an array of labels. */

  fprintf(outfile,"[ ");

  for(i=0,tickpos=tot_length;i<=NTICKS;i++,tickpos -= tick_s_step) {

    fprintf(outfile,"(%1.2f) ",tickpos);
    
  }

  fprintf(outfile," ]\n");

  /* Now display them. */

  fprintf(outfile,"aload pop \n");
  fprintf(outfile,
	  "gsave TF setfont \n"
	  "0 1 %d { pop dt %g %g translate 0 0 moveto } for grestore \n",
	  NTICKS,tickplotstep,tickplotstep);

  /* We now add tags to the curvature plot. */

  fprintf(outfile,"/lt { %g 0 moveto dup stringwidth pop neg %g rmoveto show"
	  " newpath 0 0 moveto %g 0 lineto stroke } def \n",
	  -12/plotscale,-10/plotscale,-3/plotscale); 

  fprintf(outfile,"gsave\n");
  fprintf(outfile,"tls TF setfont\n");
  fprintf(outfile,"0 0 moveto 0 %g translate\n",-2.5*tickplotstep);
  
  fprintf(outfile,"(2) (1) (0) 0 1 2 { pop lt 0 %g translate } for\n",
	  1.0*tickplotstep);

  fprintf(outfile,"grestore\n");
  
  /* Now we identify regions with no contact (straight segments). */
  /* If we're later desperate for filesize, we can come back and
     munger with this. */

  /* Our method is to search the list of struts for suspiciously large 
     gaps between s values. If such a gap is larger than 3*avg_edgelength,
     we will become suspicious that a straight segment is indicated. */

  /* lshape takes 3 arguments, slo shi end and draws an "L-shape" */

  fprintf(outfile, "/lshp { "
	  "exch    "		/* l e h */          /* Create many copies of args */
	  "dup dup "		/* l e h h h */
	  "5 2 roll "		/* h h l e h */
	  "3 -1 roll "		/* h h e h l */
	  "dup "		/* h h e h l l */
	  "4 1 roll "           /* h h l e h l */
	  "3 -1 roll "          /* h h l h l e */
	  "dup "                /* h h l h l e e */
	  "4 1 roll "           /* h h l e h l e */
	  "exch "               /* h h l e h e l */
	  "dup "                /* h h l e h e l l */
	  "4 2 roll "           /* h h l e l l h e */
	  "3 1 roll "           /* h h l e l e l h */
	  "dup dup "            /* h h l e l e l h h h */
	  "4 3 roll "           /* h h l e l e h h h l */
	  "dup dup "            /* h h l e l e h h h l l l */

	  "newpath "
	  "0 moveto "                  /* l 0 moveto */
	  "0 1 4 { pop lineto } for "
				/* l l lineto */
				/* h h lineto */
				/* e h lineto */
	                	/* e l lineto */
	  			/* h l lineto */

	  "0 lineto "          	/* 0 h lineto */ /* empty stack */
	  "closepath fill } def ");

  fprintf(outfile,"gsave \n");
  fprintf(outfile,"0.7 0.9 1 setrgbcolor\n");

  maxskip = 0; straightlength = 0;

  for(i=1;i<2*n_struts;i++) {

    s_skip = as_list[i].s - as_list[i-1].s;
    
    maxskip = (s_skip > maxskip) ? s_skip : maxskip;

    if (s_skip > 3*avg_edgelength) {

      straightlength += s_skip;

      slo = (int)(10*as_list[i-1].s/boxwidth) + 5;
      shi = (int)(10*as_list[i].s/boxwidth) - 5;

      fprintf(outfile,"%d %d %d lshp\n",slo,shi,10*NBOXES);

      fprintf(outfile,"%d %g %d %g rectfill\n",
	      slo,-2.5*tickplotstep,shi-slo,2*tickplotstep);

      /* Now add "smart" tags on the rhs. */
      
      fprintf(outfile,
	      "gsave tls TF setfont %d %d translate 0 0 "
	      "moveto %g %g lineto stroke "
	      "%g %g moveto 0 %g rmoveto (%1.2f) show grestore\n",
	      10*NBOXES,shi,
	      8.0/plotscale,6.0/plotscale,10.0/plotscale,9.0/plotscale,
	      -5/plotscale,as_list[i].s);

      fprintf(outfile,
	      "gsave tls TF setfont %d %d translate 0 0 "
	      "moveto %g %g lineto stroke "
	      "%g %g moveto 0 %g rmoveto (%1.2f) show grestore\n",
	      10*NBOXES,slo,
	      8.0/plotscale,-6.0/plotscale,10.0/plotscale,-9.0/plotscale,
	      -1/plotscale,as_list[i-1].s);    
      
    }
    
  }

  fprintf(outfile,"grestore\n");

  printf("Searched for straight segments...\n");
  
  if (straightlength > 0) {

    printf("\tFound a total of %g units of straight segment\n",
	   straightlength);

  } else {

    printf("\tNone found. Max strut s value gap: %g\n"
	   "\t            3*avg_edgelen        : %g\n",
	   s_skip,3*avg_edgelength);

  }


  /* We now do highlighting and labelling of the kinked vertices in mrloc. */

  if (num_min_rad_locs != 0) {  /* There may _be_ no kinks, so we might just skip */

    fprintf(outfile,"/ht { 0 0 moveto 0 -35 lineto stroke 0 %g moveto " /* Draw tick */
	    " dup stringwidth pop neg 2 div 0 rmoveto show "
	    " } def\n",-4 - 10.0/plotscale);
    
    fprintf(outfile,"gsave\n");
    fprintf(outfile,"1 0.2 0.2 setrgbcolor\n"); /* Kinks are highlighted in red. */ 
    
    qsort(min_rad_locs,num_min_rad_locs,sizeof(octrope_mrloc),compare_mrlocs);

    for(i=0;i<num_min_rad_locs;i++) {

      /* i is now at the start of a kink arc (run of consecutive mr locs) */

      for(j=i+1;
	  j<num_min_rad_locs && 
	    min_rad_locs[j].component == min_rad_locs[i].component && 
	    min_rad_locs[j].vert == min_rad_locs[j-1].vert + 1;
	  j++);

      j--;

      /* j is now at the end of the kink arc */

      slo=(int)(10*s_positions[min_rad_locs[i].component][min_rad_locs[i].vert]/boxwidth)-5;
      shi=(int)(10*s_positions[min_rad_locs[j].component][min_rad_locs[j].vert]/boxwidth)+5;

      /* Note that we fatten kinks a little to display one-vertex kinks with 
	 something like a responsible error bar. */

      fprintf(outfile,"%d %d %d lshp\n", /* Highlight on the main plot */
	      slo,shi,10*NBOXES);

      fprintf(outfile,"%d %g %d %g rectfill\n", /* Highlight on the curvature plot */
	      slo,-2.5*tickplotstep,shi-slo,2*tickplotstep);
      
      /* Now do horizontal tagging below the curvature plot */

      fprintf(outfile,"gsave tls TF setfont %g %g translate (%1.2f) ht grestore\n",
	      (shi+slo)/2.0,-2.5*tickplotstep,
	      (s_positions[min_rad_locs[i].component][min_rad_locs[i].vert] +
	       s_positions[min_rad_locs[j].component][min_rad_locs[j].vert])/2.0);

    }

  }

  /* Now put a nice subtle grid on the whole thing */
  
  fprintf(outfile,
	  "gsave "
	  "0.98 setgray "
	  "%g  setlinewidth \n",0.75/plotscale);

  for(i=1;i<NTICKS;i++) {
    
    fprintf(outfile,"newpath %g 0 moveto %g %g lineto %d %g lineto stroke\n",
	    i*tickplotstep,i*tickplotstep,i*tickplotstep,
	    NBOXES*10,i*tickplotstep);
  }
  
  fprintf(outfile,"grestore\n");
  
  /* Including extending vertical tick marks onto the curvature plot below */

  fprintf(outfile,
	  "gsave "
	  "%g  setlinewidth \n",0.75/plotscale);

  for(i=1;i<NTICKS;i++) {

    fprintf(outfile,
	    "0.8  setgray newpath %g 0  moveto 0 %g rlineto stroke\n"
	    "0.98 setgray newpath %g %g moveto 0 %g rlineto stroke\n",
	    i*tickplotstep,-0.5*tickplotstep,
	    i*tickplotstep,-0.5*tickplotstep,-2*tickplotstep);
  
  }

  fprintf(outfile,"0.98 setgray newpath 0 %g moveto %d 0 rlineto stroke\n",
	  -1.5*tickplotstep,10*NBOXES);

  fprintf(outfile,"grestore\n");

  /* And now we plot the struts themselves. */

  fprintf(outfile,"0.120 0.360 0.400 setrgbcolor\n"); 
  
  fprintf(outfile,"[ ");
     
  for(i=0;i<2*n_struts;i++) {
  
    if (as_list[i].t < as_list[i].s) { /* Display lower triangle. */

      fprintf(outfile, " %d %d ",
	      (int)(10*as_list[i].s/boxwidth) - 5,
	      (int)(10*as_list[i].t/boxwidth) - 5);

    }

    if (!((i+1)%20)) { fprintf(outfile,"\n"); }

  }
  
  fprintf(outfile,"] aload pop \n");
  fprintf(outfile,"0 1 %d { pop 10 10 rectfill } for \n",n_struts-1);

  /* We now actually plot the curvature data on the new box. */

  printf("Now plotting curvature data (with subsampling)...\n");

  fprintf(outfile,"gsave\n");
  fprintf(outfile,"0 0 moveto 0 %g translate\n",
	  -2.5*tickplotstep);  

  fprintf(outfile,"%g setlinewidth\n",0.5/plotscale);
  fprintf(outfile,"0.120 0.360 0.400 setrgbcolor\n"); 
  fprintf(outfile,"1 setlinejoin\n");
  fprintf(outfile,"1 setlinecap\n");
  fprintf(outfile,"0 0 %d %g rectclip\n",10*NBOXES,2*tickplotstep);

  curvpixel = 0.5/plotscale; 

  /* This is "one pixel" on the printed graph at 150 dpi * 1.5 inches */
  /* In order to compress the curvature data, we plot a point, 
     and then read ahead, skipping measurements as long as 
     they are within 2*curvpixel of the line joining the last 
     plotted point and the current address. */

  printed_total = 0;

  for(cmp=0;cmp<L->nc;cmp++) {

    printedpt = 0;

    for(vert=1;vert<L->cp[cmp].nv;vert++) { 

      intpos = (int)(10*s_positions[cmp][vert]/boxwidth);
      yi = curv = octrope_link_curvature(L,cmp,vert)*tickplotstep;

      fprintf(outfile,"%d %g ",intpos,curv);      
      printedpt++;

      if (!(printedpt%20)) { fprintf(outfile,"\n"); }
      if (vert > L->cp[cmp].nv-3) { continue; }

      /* Now we attempt to read ahead */

      for(pt_ok=TRUE,j=vert+2;pt_ok && j < L->cp[cmp].nv;j++) {

	yj = octrope_link_curvature(L,cmp,j)*tickplotstep;
	mij = (yj - yi)/(s_positions[cmp][j] - s_positions[cmp][vert]);

	/* We must check every point in i+1..j-1 for closeness to the line */

	for(k=vert+1;k<j;k++) {

	  predicted_curvk = mij * (s_positions[cmp][k] - s_positions[cmp][vert]);
	  
	  if (fabs(predicted_curvk - tickplotstep*octrope_link_curvature(L,cmp,k)) > 2*curvpixel) {
	    
	    pt_ok = FALSE;

	  }

	}

      }

      /* We have reached the point where i and j-2 are the farthest apart we can get */
      /* without causing more than 2*curvpixel of error in the plot. We set i to j-3 */
      /* since it will be incremented when we return to the top of the loop. */
      
      vert = j-3;
      
    }

    fprintf(outfile,"moveto 2 1 %d { pop lineto } for stroke\n",
	    printedpt - 1);

    printed_total += printedpt;

  }

  fprintf(outfile,"grestore\n");
  printf("\t Printed %d points (of %d possible).\n",printed_total,octrope_link_edges(L));

  /* And finally frame the whole thing */

  fprintf(outfile,
	  "0 0 moveto %d 0 rlineto "
	  "0 %d rlineto -%d -%d rlineto "
	  "0.6 setgray stroke \n",
	  NBOXES*10,NBOXES*10,NBOXES*10,NBOXES*10);

  fprintf(outfile,"0.6 setgray 0 %g %d %g rectstroke",
	  -2.5*tickplotstep,10*NBOXES,2*tickplotstep);
	  
  /* We're done! Free memory and quit. */
  
  free(as_list);
  free(s_positions);  
    
}
        
int compare_edgerecs(const void *a, const void *b)
{
  struct edge_record *A,*B;

  A = (struct edge_record *)(a);
  B = (struct edge_record *)(b);

  if (A->n_struts - B->n_struts != 0) {

    return B->n_struts - A->n_struts;

  } 

  if (A->comp - B->comp != 0) {

    return A->comp - B->comp;

  }

  return A->lead_vert - B->lead_vert;

}

void analyze_and_print_struts(octrope_link *L, int n_struts, 
                              octrope_strut *strutlist)

{
  struct edge_record *edgedata;
  int    *offsets;

  int    i,j;

  fprintf(stderr,"\nstruts: Building edge database.\n");

  /* We first build a table of edge records. */

  offsets    =  (int *)(calloc(L->nc,sizeof(int)));
  offsets[0] = 0;
  
  for (i=1;i<L->nc;i++) {
    offsets[i] = pline_edges(L->cp[i-1]) + offsets[i-1];
  }
    
  edgedata = (struct edge_record *)(calloc(octrope_link_edges(L),
					   sizeof(struct edge_record)));

  for(i=0; i < L->nc; i++) {
    for(j=0; j < pline_edges(L->cp[i]); j++) {
      edgedata[offsets[i] + j].comp = i;
      edgedata[offsets[i] + j].lead_vert = j;
      edgedata[offsets[i] + j].n_struts = 0;
    }
  }

  /* Now we go through and count the number of struts in which each edge is *
   * involved.                                                              */

  for(i=0;i<n_struts;i++) {
    edgedata[offsets[strutlist[i].component[0]] + strutlist[i].lead_vert[0]].n_struts++;
    edgedata[offsets[strutlist[i].component[1]] + strutlist[i].lead_vert[1]].n_struts++;
  }

  /* Sort by that count ... */
  fprintf(stderr,"struts: Sorting %d edges by number of incident struts (%d struts total)...\n\n",
          octrope_link_edges(L),n_struts);

  qsort(edgedata,octrope_link_edges(L),sizeof(struct edge_record),compare_edgerecs);

  /* And print out the results. */

  for(i=0; i< min(octrope_link_edges(L),10);i++) {
    printf("Comp: %d  Lead Vert: %d  # of struts: %d.\n",
           edgedata[i].comp,edgedata[i].lead_vert,edgedata[i].n_struts);
  }

  /* Now we free some of our memory. */

  free(offsets);
  free(edgedata);
}

int main(int argc,char *argv[]) {

  FILE *infile_fptr,*outfile_fptr;
  octrope_link *L;

  int    sl_size,n_struts,s_cnt;
  octrope_strut *strutlist;
  octrope_mrloc *mrloc_list;
  int num_mr_locs,mr_size;
  double shortest;
  double eps = 0.01;
  
  int nerrors;
  char outfile_name[1000];
  
  struct arg_file *infile = arg_file1(NULL,NULL,"<file>", "input file");
  struct arg_int  *levels = arg_int0("l", "levels","<n>", "number of octree levels");
  struct arg_int  *debuglevel = arg_int0("v", "verbosity","<n>", "level of debugging information to print");
  struct arg_lit  *stats  = arg_lit0("s","stats","display strut set statistics");
  struct arg_dbl  *epsilon= arg_dbl0("e","epsilon","<x>","find struts within this tolerance of minimum length");
  struct arg_lit  *show   = arg_lit0("S","show","Show struts");
  struct arg_lit  *help   = arg_lit0("h","help","display help message");
  struct arg_lit  *no_file= arg_lit0("n","nofile","Don't write out the strut file");
  struct arg_lit  *stplot = arg_lit0(NULL,"stplot","Create a (PostScript) plot of strut locations.");
  struct arg_end  *end = arg_end(20);

  void *argtable[] = {help,epsilon,levels,stats,show,no_file,stplot,debuglevel,infile,end};
  
  struct arg_end  *helpend = arg_end(20);

  void *helptable[] = {help, helpend};
  
  /* We start by parsing the command-line arguments with argtable. */
  
  if (arg_nullcheck(argtable) != 0 || arg_nullcheck(helptable) != 0)
    printf("error: insufficient memory\n");

  nerrors = arg_parse(argc,argv,argtable);
  
  if (nerrors > 0) { /* The standard syntax didn't match-- try the help table */

    nerrors = arg_parse(argc,argv,helptable);

    if (nerrors > 0) {  /* The help table didn't match either-- the first set of */
                        /* errors was probably more helpful, so we display it. */

      arg_print_errors(stdout,helpend,"helptable");
      arg_print_errors(stdout,end,"struts");
      exit(1);

    } else {  /* The help table matched, which means we asked for help or gave nothing */
  
      printf("struts computes the set of self-contacts of a Geomview VECT file.\n"
           "usage: \n\n");
      arg_print_glossary(stdout, argtable," %-25s %s\n");
      exit(0);

    }
    
  }
  
  /* We have now parsed the command-line args, and we slot them into the proper globals. */
  
  if (levels->count > 0) {
    
    octrope_set_levels(*(levels->ival));
    
  }
  
  if (debuglevel->count > 0) {
    
    octrope_set_debug(*(debuglevel->ival));
    printf("Debug level set to %d.\n",*(debuglevel->ival));
    
  }

  if (epsilon->count > 0) {

    eps = *(epsilon->dval);

  }
  
  /* Next, we mess around with opening files and such. */
  /* Notice that "argtable" will have already objected if no filename was passed */
  
  infile_fptr = fopen(*(infile->filename),"r");
  
  if (infile_fptr == NULL) {
    
    fprintf(stderr,"struts: Couldn't open file %s.\n",*(infile->filename));
    exit(0);

  }
  
  L = octrope_link_read(infile_fptr);
  
  /* We now demonstrate the octrope library's error handling protocol: */
  
  if (octrope_error_num > 0) {   /* This is the signal for an error. */
    
    fprintf(stderr,"struts: link reading error\n%s\n",octrope_error_str);
    exit(octrope_error_num);
    
  }
  
  fclose(infile_fptr);
  
  /* We now compute the strut set. */
  
  sl_size = 20*octrope_link_edges(L);      /* We allocate an extra-large buffer. */
  strutlist = (octrope_strut *)(calloc(sl_size,sizeof(octrope_strut)));
  
  n_struts = octrope_struts(L,0,eps,strutlist,sl_size,&shortest,NULL,0);
  
  if (octrope_error_num > 0) {
    
    fprintf(stderr,"%s\n",octrope_error_str);
    exit(1);
    
  }

  /* We now compute the set of minrad locations */

  mr_size = 2*octrope_link_edges(L);      /* We allocate an extra-large buffer. */
  mrloc_list = (octrope_mrloc *)(calloc(mr_size,sizeof(octrope_mrloc)));
  
  octrope_minrad(L,0,eps,mrloc_list,mr_size,&num_mr_locs);
  
  if (octrope_error_num > 0) {
    
    fprintf(stderr,"%s\n",octrope_error_str);
    exit(1);
    
  }
  
  /* We now go ahead and check whether we should output human-readable data. */
  
  if (stats->count > 0) {
    
    analyze_and_print_struts(L,n_struts,strutlist);
    
  }   
  
  if (show->count > 0) {
    for (s_cnt = 0; s_cnt < n_struts; s_cnt++) {
      printf("%d:%d - %d:%d %15.15f %15.15f\n",
        strutlist[s_cnt].component[0],strutlist[s_cnt].lead_vert[0],
        strutlist[s_cnt].component[1],strutlist[s_cnt].lead_vert[1],
        strutlist[s_cnt].position[0],strutlist[s_cnt].position[1]
      );
    }
  }

  /* We check for postscript output. */

  if (stplot->count > 0) { 
  
     if (strlen(*(infile->basename)) > sizeof(outfile_name)-20) {
      
        fprintf(stderr,"struts: Ridiculously long input filename can't be parsed.\n");
        exit(1);
      
      }
    
      sprintf(outfile_name,"%s",*(infile->basename));
    
      if (strstr(outfile_name,".vect") != NULL) {
  
        sprintf(strstr(outfile_name,".vect"),".stplot.ps");
  
       } else {
  
         strcat(outfile_name,".stplot.ps");
  
       }
    
       outfile_fptr = fopen(outfile_name,"w");
    
       if (outfile_fptr == NULL) {
      
         fprintf(stderr,"struts: Couldn't open %s for writing.\n",outfile_name);
         exit(1);
       
        }

       create_st_plot(L,n_struts,strutlist,num_mr_locs,mrloc_list,outfile_fptr);       
       fclose(outfile_fptr);
    
       printf("Wrote s-t plot to file %s.\n",outfile_name); 
   }    
  
  /* We now prepare to output the data to a SKEL file. */   
  
  if (no_file->count == 0) {
  
    if (strlen(*(infile->basename)) > sizeof(outfile_name)-20) {
      
      fprintf(stderr,"struts: Ridiculously long input filename can't be parsed.\n");
      exit(1);
      
    }
    
    /* Create filename for SKEL output. */
    
    sprintf(outfile_name,"%s",*(infile->basename));
    
    if (strstr(outfile_name,".vect") != NULL) {
  
      sprintf(strstr(outfile_name,".vect"),".struts.skel");
  
    } else {
  
      strcat(outfile_name,".struts.skel");
  
    }
    
    outfile_fptr = fopen(outfile_name,"w");
    
    if (outfile_fptr == NULL) {
      
      fprintf(stderr,"struts: Couldn't open %s for writing.\n",outfile_name);
      exit(1);
      
    }
    
    /* Now we write the strut set to a Geomview "SKEL" data structure. */
    
    octrope_vector se[2];
    int i;
    
    fprintf(outfile_fptr,"SKEL\n");
    fprintf(outfile_fptr,"%d %d \n",2*n_struts,n_struts);
    
    for(i=0;i<n_struts;i++) {
      
      octrope_strut_ends(L,&strutlist[i],se);
      fprintf(outfile_fptr,"%g %g %g \n",se[0].c[0],se[0].c[1],se[0].c[2]);
      fprintf(outfile_fptr,"%g %g %g \n",se[1].c[0],se[1].c[1],se[1].c[2]);
      
    }
    
    for(i=0;i<n_struts;i++) {
      
      fprintf(outfile_fptr,"2 %d %d \n",2*i,2*i+1);
      
    }
    
    /* Now we close files and free memory. */
    
    printf("struts: %d struts written to %s.\n",n_struts,outfile_name);
    fclose(outfile_fptr);   
      
  } else {
    printf("struts: %d struts found.\n",n_struts);
  }
  free(strutlist);
  arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));

  exit(0);
}
